Unit LoadPCX;

{=======================================================================}
{ Unit LoadPCX by AX                                                  }
{**********************                                                 }
{                                                                       }
{ Support PCX version 5.0, 8 bits per pixel, palette at the end of file }
{ and gray scale palette. Uses disk read buffering.                     }
{-----------------------------------------------------------------------}
{ function loadpcx_pal(fname:string; pal:pointer): byte;                }
{ Loads palette (768 bytes) from PCX file into pal^. Palette colors     }
{ are in range 0..63. If grayscale PCX - generates palette.             }
{ 0 - ok, 1 - can't open file, 2 - invalid PCX                          }
{                                                                       }
{-----------------------------------------------------------------------}
{ function loadpcx_raw(fname:string; raw:pointer): byte;                }
{ Loads raw data ((x2-x1+1)*(y2-y1+1) bytes) from PCX file into raw^.   }
{ 0 - ok, 1 - can't open file, 2 - invalid PCX                          }
{                                                                       }
{-----------------------------------------------------------------------}
{ function loadpcx_raw_pal(fname:string; raw,pal:pointer): byte;        }
{ Do the same loadpcx_raw&loadpcx_pal.                                  }
{                                                                       }
{-----------------------------------------------------------------------}
{ function loadpcx_info(fname:string; header: pcxheader): byte;         }
{ Loads pcxheader from file.                                            }
{ 0 - ok, 1 - can't open file, 2 - invalid PCX                          }
{                                                                       }
{=======================================================================}

INTERFACE

type
{   PCX }
 PCXHEADER = record
     manuf      :byte;            { ᥣ =10  Paintbrush      }
     hard       :byte;            { ଠ  ᨨ            }
     encod      :byte;            { 㯯 ஢(=1)      }
     bitpx      :byte;            {                      }
     x1         :word;            {  ⨭(.)     }
     y1         :word;            {                                }
     x2         :word;            {                                }
     y2         :word;            {                                }
     hres       :word;            { ਧ.ࠧ襭 ᯫ       }
     vres       :word;            { ⨪.ࠧ襭 ᯫ      }
     clrma      :array[0..47] of byte;  {                     }
     vmode      :byte;            { ()                }
     nplanes    :byte;            { - ᪮⥩ (.2.5=0) }
     bplin      :word;            {   ப                }
     palinfo    :word;            { .  (1=.,2=)}
     shres      :word;            { 襭 ᪠           }
     svres      :word;
     xtra       :array[0..53] of byte; { .⮥  (䨫)  }
 end;

function loadpcx_pal(fname:string; pal:pointer): byte;
function loadpcx_raw(fname:string; raw:pointer): byte;
function loadpcx_raw_pal(fname:string; raw,pal:pointer): byte;
function loadpcx_info(fname:string; header:pcxheader): byte;

IMPLEMENTATION

var
 dummy: byte;

function open_and_check: byte; assembler;
{ ===============================}
{ IN: es:si = fname:string       }
{ OUT : bx - file hanlde         }
{       AL=0 - ok                }
{       AL=1 - can't open        }
{       AL=2 - invalid PCX       }
{================================}
var
 s: array [1..256] of byte;
asm
 {=== Convert to ASCIIZ ===}
 push es
 pop ds
 mov cl,[si]
 xor ch,ch
 jcxz @cantopen
 inc si
 push ss
 pop es
 lea di,s
 cld
 rep movsb
 xor al,al
 stosb

 {=== Open File ===}
 mov ax,3d00h
 push ss
 pop ds
 lea dx,s
 int 21h
 jc @cantopen

 mov bx,ax

 {=== read 4 bytes ===}
 mov cx,4
 lea dx,s
 mov ah,3fh
 int 21h
 jc @badpcx

 cmp word ptr s,050ah    {must be  = 0A 05 01 08 =}
 jne @badpcx
 cmp word ptr s+2,0801h
 jne @badpcx

 xor al,al
 jmp @exit

@badpcx:
 mov ah,3eh
 int 21h
 mov al,2
 jmp @exit

@cantopen:
 mov al,1

@exit:
 mov cx,seg dummy
 mov ds,cx
end;

function readpal: byte; assembler;
{===========================}
{IN:  es:dx - pointer to buf}
{     bx - file handle      }
{OUT: al=1 - error          }
{===========================}
var
 tmp_byte : byte;
asm
 push es
 push dx

{=== Seek to palinfo ===}
 mov ax,4200h
 xor cx,cx
 mov dx,43h
 int 21h
 jc @error1

{=== Read palinfo byte ===}
 mov cx,1
 push ss
 pop ds
 lea dx,tmp_byte
 mov cx,1
 mov ah,3fh
 int 21h
 jc @error1

 cmp tmp_byte,1
 je @color_pal

{=== Generate Gray palette ===}
 pop di
 pop es

 mov cx,256
 xor ah,ah
 cld

 @l1:
 mov al,ah
 shr al,1
 shr al,1
 stosb
 stosb
 stosb
 inc ah
 loop @l1
 xor al,al
 jmp @exit

@color_pal:
 {=== Seek to palette ===}
 mov ax,4202h
 mov cx,-1
 mov dx,-769
 int 21h
 jc @error1

{=== read palettetype info ===}
 push ss
 pop ds
 lea dx,tmp_byte
 mov cx,1
 mov ah,3fh
 int 21h
 jc @error1

 mov al,tmp_byte
 cmp al,10
 je @ok_pal
 cmp al,0ch
 je @ok_pal
 jmp @error1

@ok_pal:

{=== Read palette ===}
 pop dx
 pop ds

 mov cx,768
 mov ah,3fh
 int 21h
 jc @error

 cmp tmp_byte,0ch  {== must shl ===}
 jne @l2

 mov si,dx
 mov cx,768

 @l3:
 mov al,[si]
 shr al,1
 shr al,1
 mov [si],al
 inc si
 loop @l3

 @l2:
 xor al,al
 jmp @exit

@error:
 mov al,1
 jmp @exit

@error1:
 add sp,4
 mov al,1

@exit:
 mov cx,seg dummy
 mov ds,cx
end;

procedure readraw; assembler;
{=====================}
{ IN:  ES:DI - BUF    }
{      BX - handle    }
{ OUT: al = 1 - error }
{=====================}
var
 tmp_byte,tmp_byte1: byte;
 count: array [1..8] of byte;
 inbuf: word;
 buf : array [1..2048] of byte;
 bufpoi: word;
asm

{=== Seek to size ===}
 mov ax,4200h
 xor cx,cx
 mov dx,4
 int 21h
 jc @error

{=== Read size info}
 mov ah,3fh
 push ss
 pop ds
 lea dx,count
 mov cx,8
 int 21h
 jc  @error

{=== Calculate counter ===}
 mov ax,word ptr count+4
 sub ax,word ptr count
 inc ax
 mov cx,word ptr count+6
 sub cx,word ptr count+2
 inc cx
 mul cx
 mov word ptr count,ax
 mov word ptr count+2,dx

{=== Seek to raw data ===}
 mov ax,4200h
 xor cx,cx
 mov dx,128
 int 21h
 jc @error

 push ss
 pop ds
 cld
 mov inbuf,0

@cont:
{=== Read one byte ===}
 call @read_byte_buf

 mov cl,al
 cmp al,192    { if byte > 192 - do RLE decoding }
 jnb @do_rle

 mov cx,1
 jmp @cont1

@do_rle:

{=== Read one byte ===}
 call @read_byte_buf

{=== decode ===}
 xor ch,ch
 sub cl,192

@cont1:
 mov dx,cx
{$IFNDEF DPMI}
 rep stosb

{=== Normalize pointer ===}
@l2:
 cmp di,16
 jna @l1
 sub di,16
 mov ax,es
 inc ax
 mov es,ax
 jmp @l2
 @l1:
{$ELSE}
 @lp2:
 cmp di,0ffffh
 jne @lp1

 stosb
 mov ax,es
 add ax,selectorinc
 mov es,ax
 @lp1:

 stosb
 loop @lp2


{$ENDIF}

{=== Decrement Counter ===}
 xor ax,ax
 sub word ptr count,dx
 sbb word ptr count+2,ax
 jc @ok
 cmp word ptr count,ax
 jne @cont
 cmp word ptr count+2,ax
 jne @cont

@ok:
 xor al,al
 jmp @exit
@error:
 mov al,1
 jmp @exit

@read_byte_buf:
 cmp inbuf,0
 je @l5

@l6:
 lodsb
 dec inbuf
 ret

@l5:
 lea si,buf
 push cx
 push si
 mov cx,2048
 mov ah,3fh
 lea dx,buf
 int 21h
 pop si
 pop cx
 jc @error
 mov inbuf,ax
 or ax,ax
 jz @error
 cld
 jmp @l6


@exit:
 mov cx,seg dummy
 mov ds,cx
end;

function loadpcx_raw_pal(fname:string; raw,pal:pointer): byte; assembler;
{=========================}
{ OUT:0 - ok              }
{     1 - can't open      }
{     2 - invalid PCX     }
{=========================}

asm
 {=== Open&Check ===}
 les si,fname
 call open_and_check
 or al,al
 jnz @exit

 {=== Read Palette ===}
 les dx,pal
 call readpal
 or al,al
 jnz @error1

 {=== Read Raw Data ===}
 les di,raw
 call readraw
 or al,al
 jnz @error1

{=== Close File ===}
 mov ah,3eh
 int 21h
 xor al,al
 jmp @exit

@error1:
 mov ah,3eh
 int 21h
 mov al,2

@exit:
 mov cx,seg dummy
 mov ds,cx
end;

function loadpcx_raw(fname:string; raw:pointer): byte; assembler;
{=========================}
{ OUT:0 - ok              }
{     1 - can't open      }
{     2 - invalid PCX     }
{=========================}

asm
 {=== Open&Check ===}
 les si,fname
 call open_and_check
 or al,al
 jnz @exit

 {=== Read Raw Data ===}
 les di,raw
 call readraw
 or al,al
 jnz @error1

{=== Close File ===}
 mov ah,3eh
 int 21h
 xor al,al
 jmp @exit

@error1:
 mov ah,3eh
 int 21h
 mov al,2

@exit:
 mov cx,seg dummy
 mov ds,cx
end;

function loadpcx_pal(fname:string; pal:pointer): byte; assembler;
{=========================}
{ OUT:0 - ok              }
{     1 - can't open      }
{     2 - invalid PCX     }
{=========================}

asm
 {=== Open&Check ===}
 les si,fname
 call open_and_check
 or al,al
 jnz @exit

 {=== Read Palette ===}
 les dx,pal
 call readpal
 or al,al
 jnz @error1

{=== Close File ===}
 mov ah,3eh
 int 21h
 xor al,al
 jmp @exit

@error1:
 mov ah,3eh
 int 21h
 mov al,2

@exit:
 mov cx,seg dummy
 mov ds,cx
end;

function loadpcx_info(fname:string; header:pcxheader): byte; assembler;
{=========================}
{ OUT:0 - ok              }
{     1 - can't open      }
{     2 - invalid PCX     }
{=========================}

asm
 {=== Open&Check ===}
 les si,fname
 call open_and_check
 or al,al
 jnz @exit

{=== seek to 0:0 ===}
 mov ax,4200h
 xor cx,cx
 xor dx,dx
 int 21h
 jc @error1

{=== load 128 bytes info ===}

 lds dx,header
 mov cx,128
 mov ah,3fh
 int 21h
 jc @error1
 cmp al,128
 jc @error1

{=== Close File ===}
 mov ah,3eh
 int 21h
 xor al,al
 jmp @exit

@error1:
 mov ah,3eh
 int 21h
 mov al,2

@exit:
 mov cx,seg dummy
 mov ds,cx
end;

begin
end.