{
=================================================================
 File structure:
  npoints     : word;
  new_npoints : word;
  points      : array [1..new_npoints,1..3] of integer;
  cross_ref   : array [1..npoints] of word;  (=1..new_npoints)
  normals_p   : array [1..npoints,1..3] of integer;
=================================================================
}

uses crt;

const
 infile  = 'laga';
 outfile = 'laga.3pk';

 maxnpoints = 3500;
 maxnpolys = maxnpoints div 4;
 maxnnewpoints = 1000;

var
 points3d    : array [1..maxnpoints,1..3] of integer;
 poly3d      : array [1..maxnpolys,1..4,1..3] of integer absolute points3d;
 new_points  : array [1..maxnnewpoints,1..3] of integer;
 cross_ref   : array [1..maxnpoints] of word;
 s           : string;
 npoints     : word;
 new_npoints : word;
 npolys      : word;
 f           : file;
 i,j         : integer;
 poly_3d     : array [1..4,1..3] of integer;
 normals     : array [1..maxnpolys,1..3] of integer;
 normals_p   : array [1..maxnpolys,1..4,1..3] of integer;
 buf : array [1..4,1..3] of integer;

 label l1;


procedure calcnormal(norm: pointer);
type
 normalt = array [1..3] of integer;
var
 x1,y1,z1   : integer;
 x21,x31, y21,y31, z21,z31    : longint;
 a,b,c,h,m  : real;
 i,j : word;

begin
 x1:=poly_3d[1,1];
 y1:=poly_3d[1,2];
 z1:=poly_3d[1,3];

 x21:=poly_3d[2,1]-x1;
 y21:=poly_3d[2,2]-y1;
 z21:=poly_3d[2,3]-z1;

 x31:=poly_3d[3,1]-x1;
 y31:=poly_3d[3,2]-y1;
 z31:=poly_3d[3,3]-z1;

 a:= y21*z31 - y31*z21;
 b:= z21*x31 - x21*z31;
 c:= x21*y31 - x31*y21;

 m:=sqrt(sqr(a)+sqr(b)+sqr(c));
{if m<>0 then
 begin}
 a:=a*64/m;
 b:=b*64/m;
 c:=c*64/m;
{end;}
 normalt(norm^)[1]:=round(a);
 normalt(norm^)[2]:=round(b);
 normalt(norm^)[3]:=round(c);
end;

procedure calcnormals(n:word; polys: pointer);
type
 polyst = array [1..1000,1..4,1..3] of integer;
var
 i: integer;
begin
 for i:=1 to n do
   begin
    write(i,#13);
     move(addr(polyst(polys^)[i])^,addr(poly_3d)^,4*3*2);
     calcnormal(addr(normals[i]));
   end;
end;

procedure calcnormals_p;
var
 a,b,c : longint;
 n     : byte;
 i,j,k,m  : integer;

begin
 for i:=1 to npolys do
    for j:=1 to 4 do
      begin
      write(i,#13);
      a:=0;  b:=0;  c:=0; n:=0;

        for k:=1 to npolys do
          for m:=1 to 4 do
           if ((poly3d[i,j,1]=poly3d[k,m,1]) and
              (poly3d[i,j,2]=poly3d[k,m,2]) and
              (poly3d[i,j,3]=poly3d[k,m,3])) then
            begin
             inc(a,normals[k,1]);
             inc(b,normals[k,2]);
             inc(c,normals[k,3]);
             inc(n);
            end;
        normals_p[i,j,1]:=a div n;
        normals_p[i,j,2]:=b div n;
        normals_p[i,j,3]:=c div n;
      end;
end;

     label l2;
begin
clrscr;
 assign(f,infile); reset(f,1);
 blockread(f,s,15);
 blockread(f,npoints,2);
 dec(npoints,2);
 blockread(f,points3d,6+6);
 blockread(f,points3d,npoints*6);
 close(f);
 npolys:=npoints div 4;

 writeln('Input Filename: ',infile);
 writeln('Number of points: ',npoints);
 writeln('Number of polys: ',npolys);
 writeln;

writeln('Veryfing polys...');

for i:=1 to npolys do
 begin
  l2:
  write(i,#13);
  if ((poly3d[i,1,1]=poly3d[i,2,1]) and
     (poly3d[i,1,2]=poly3d[i,2,2]) and
     (poly3d[i,1,3]=poly3d[i,2,3])) or

     ((poly3d[i,2,1]=poly3d[i,3,1]) and
     (poly3d[i,2,2]=poly3d[i,3,2]) and
     (poly3d[i,2,3]=poly3d[i,3,3])) or

     ((poly3d[i,1,1]=poly3d[i,3,1]) and
     (poly3d[i,1,2]=poly3d[i,3,2]) and
     (poly3d[i,1,3]=poly3d[i,3,3])) then
         begin
          move(poly3d[i],buf,4*3*2);
          move(buf[1],poly3d[i,2],3*3*2);
          move(buf[4],poly3d[i,1],3*2);
          goto l2;
         end;
 end;

writeln('packing points...');

 new_npoints:=0;
 for i:=1 to npoints do
   begin
    write(i,#13);
     for j:=1 to new_npoints do
       if (points3d[i,1]=new_points[j,1]) and
          (points3d[i,2]=new_points[j,2]) and
          (points3d[i,3]=new_points[j,3]) then
            begin
             cross_ref[i]:=j;
             goto l1;
            end;
     inc(new_npoints);
     new_points[new_npoints,1]:=points3d[i,1];
     new_points[new_npoints,2]:=points3d[i,2];
     new_points[new_npoints,3]:=points3d[i,3];
     cross_ref[i]:=new_npoints;
     l1:
   end;

writeln('calculating polys normals...');

calcnormals(npolys,addr(points3d));

writeln('calculating points normals...');

calcnormals_p;
writeln;

writeln('Output filename: ',outfile);
writeln('Number of points: ',new_npoints);

assign(f,outfile); rewrite(f,1);
blockwrite(f,npoints,2);
blockwrite(f,new_npoints,2);
blockwrite(f,new_points,new_npoints*6);
blockwrite(f,cross_ref,npoints*2);
blockwrite(f,normals_p,npoints*6);
close(f);
end.